import logging
import yaml
from typing import Tuple

from ladm.main.generator_context import GeneratorContext
from ladm.main.generator_strategy import LanguageGeneratorStrategy, SystemGeneratorStrategy, DataGeneratorStrategy
from ladm.main.util.custom_exceptions import LADMError
from ladm.main.util.types import LanguageDependenciesOutput, SystemDependenciesOutput, DataDependenciesOutput, \
    YamlContent


class LADM:
    logger = logging.getLogger('LADM')

    def __init__(self, input_filepath: str) -> None:
        """
        :param input_filepath: File path and name of the skill.yml file containing the dependencies
                               (Or any other file containing dependencies in the correct format)
        """
        self._input_filepath: str = input_filepath
        self._language: str = ""
        self._version: str = ""

    @property
    def language(self):
        if not self._language:
            self.logger.warning("Language not set, generation must be run first")
        return self._language

    @property
    def version(self):
        if not self._version:
            self.logger.warning("Version not set, generation must be run first")
        return self._version

    def generate_dependencies_from_file(self,
                                        do_language: bool = True,
                                        do_system: bool = True,
                                        do_data: bool = True) -> Tuple[LanguageDependenciesOutput,
                                                                       SystemDependenciesOutput,
                                                                       DataDependenciesOutput]:
        """
        Main method to call from this class. Generates three lists.
        :param do_language: generate language libraries dependencies
        :param do_system: generate system dependencies
        :param do_data: generate data dependencies
        :return: Language library dependencies, system-level dependencies, data dependencies
        """
        self.logger.info(f"Generating with args{list(vars().items())[1:]}")

        dependencies_yaml = self._read_yaml()

        self._set_language_and_version(dependencies_yaml)

        if 'dependencies' not in dependencies_yaml:
            self.logger.warning("'dependencies' is not specified in skill.yml.")
            self.logger.warning("No additional dependencies will be installed.")
            return [], [], []

        language_libraries = dependencies_yaml['dependencies'].get('libraries')
        system_dependencies = dependencies_yaml['dependencies'].get('system')
        data = dependencies_yaml['dependencies'].get('data')

        language_libraries_results: LanguageDependenciesOutput = []
        system_dependencies_results: SystemDependenciesOutput = []
        data_results: DataDependenciesOutput = []

        context = GeneratorContext()
        if do_language:
            context.strategy = LanguageGeneratorStrategy(self._language)
            language_libraries_results = context.execute_strategy(language_libraries)
        if do_system:
            context.strategy = SystemGeneratorStrategy()
            system_dependencies_results = context.execute_strategy(system_dependencies)
        if do_data:
            context.strategy = DataGeneratorStrategy()
            data_results = context.execute_strategy(data)

        return language_libraries_results, system_dependencies_results, data_results

    def _read_yaml(self) -> YamlContent:
        with open(self._input_filepath, 'r') as stream:
            try:
                dependencies_yaml = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                self.logger.error(exc)
                raise LADMError(f"The supplied YAML in {self._input_filepath} cannot be parsed.")
            return dependencies_yaml

    def _set_language_and_version(self, dependencies_yaml: YamlContent) -> None:
        language_and_version = dependencies_yaml.get('language')
        if not language_and_version:
            # This should not even happen as validation is done beforehand
            raise LADMError("No language is specified in the skill.yml.")
        self._language, self._version = self._get_language_and_version(language_and_version)

    @staticmethod
    def _get_language_and_version(language_and_version: str) -> Tuple[str, str]:
        lv = language_and_version.split('-', 1)
        return lv[0].lower(), lv[1]



import logging
from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Union

from ladm.main.util.custom_exceptions import LADMError
from ladm.main.util.types import DependenciesOutput, DependenciesInput, SystemDependenciesOutput, \
    LanguageDependenciesOutput, DataDependenciesOutput, SystemDependenciesInput, LanguageDependenciesInput, \
    DataDependenciesInput


class GeneratorStrategy(ABC):
    @abstractmethod
    def generate(self, dependencies: DependenciesInput) -> DependenciesOutput:
        pass

    @staticmethod
    def convert_yaml_deps_to_dict(deps: Union[LanguageDependenciesInput,
                                              SystemDependenciesInput]) -> Dict[str, Optional[str]]:
        """
        Goes through the list of dependencies where each entry is a list containing a dependency name and a possible
        version as a second entry, converts it to a dict where the keys are the dependency names and possible
        values are the version

        :param deps: Parsed YAML input of either the system or language dependencies part of dependencies in skill.yml
        :return: dict where keys are names of packages or libraries, and optional values are versions
        """
        return {dep[0]: dep[1] if (len(dep) > 1) else None for dep in deps}

    @staticmethod
    def check_if_dependencies_empty(dependencies: DependenciesInput) -> bool:
        return dependencies == [] or dependencies is None


class SystemGeneratorStrategy(GeneratorStrategy):
    def generate(self, dependencies: SystemDependenciesInput) -> SystemDependenciesOutput:
        if self.check_if_dependencies_empty(dependencies):
            return []

        # Parse dependencies and add them to list
        deps_and_versions: dict[str, Optional[str]] = self.convert_yaml_deps_to_dict(dependencies)
        return [f"{k}={v}" if v else k for k, v in deps_and_versions.items()]


class LanguageGeneratorStrategy(GeneratorStrategy):
    def __init__(self, language: str = None) -> None:
        self._language: str = language

    def generate(self, dependencies: LanguageDependenciesInput) -> LanguageDependenciesOutput:
        """
        :return: list of language libraries as strings possibly including version numbers
        """
        if self.check_if_dependencies_empty(dependencies):
            return []

        language_libraries_generators = {
            "python": PythonGeneratorStrategy(),
            "java": JavaGeneratorStrategy()
        }

        if self._language in language_libraries_generators:
            return language_libraries_generators[self._language].generate(dependencies)
        else:
            raise LADMError(f"Language {self._language} is not supported.")


class PythonGeneratorStrategy(LanguageGeneratorStrategy):
    def generate(self, dependencies: LanguageDependenciesInput) -> LanguageDependenciesOutput:
        """
        :return: list of Python libraries as strings possibly including version numbers
        """
        # Parse libraries and add them to list
        libs_and_versions: Dict[str, Optional[str]] = self.convert_yaml_deps_to_dict(dependencies)

        # Python requirements txt needs == instead of single =.
        # (Others too, but that's not implemented) https://www.python.org/dev/peps/pep-0440/#version-specifiers
        return [f"{k}=={v}" if v else k for k, v in libs_and_versions.items()]


class JavaGeneratorStrategy(LanguageGeneratorStrategy):
    logger = logging.getLogger('JavaGeneratorStrategy')

    def generate(self, dependencies: LanguageDependenciesInput) -> LanguageDependenciesOutput:
        """
        :return: list of Java libraries as strings possibly including version numbers
        """
        self.logger.warning("LADM for Java/Gradle libraries is not yet implemented.")
        self.logger.warning(f"Language libraries {dependencies} are discarded from the process.")
        return []


class DataGeneratorStrategy(GeneratorStrategy):
    def generate(self, dependencies: DataDependenciesInput) -> DataDependenciesOutput:
        """
        :return: list of data as strings in GAIA URI format
        """
        if self.check_if_dependencies_empty(dependencies):
            return []

        return dependencies



import shutil
import unittest
from pathlib import Path
from ladm.main.main import LADM
from ladm.main.util.custom_exceptions import LADMError


class LADMTestCase(unittest.TestCase):
    default_output_path = "test-output/LADM"
    default_output_file_name = "LADM-requirements.txt"

    def tearDown(self) -> None:
        if Path(self.default_output_path).exists():
            shutil.rmtree(self.default_output_path.split('/')[0])

    def test_correct_generation(self) -> None:
        dependencies_yml_path = "resources/correct-skill-ladm/skill_with_dependencies.yml"
        ladm = LADM(dependencies_yml_path)
        results = ladm.generate_dependencies_from_file()

        self.assertEqual((["zupline==1.0.0", "tensorblow"],
                          ["libffi-dev", "default-jre=2:1.11-71"],
                          []), results)
        self.assertEqual(ladm.language, "python")
        self.assertEqual(ladm.version, "3.8.8")

    def test_no_dependencies(self) -> None:
        ladm = LADM("resources/correct-skill-ladm/skill_without_dependencies.yml")
        results = ladm.generate_dependencies_from_file()
        self.assertEqual(([], [], []), results)

    def test_empty_dependencies(self) -> None:
        ladm = LADM("resources/correct-skill-ladm/skill_with_dependencies_empty.yml")
        results = ladm.generate_dependencies_from_file()
        self.assertEqual(([], [], []), results)

    def test_system_dependencies_only(self) -> None:
        ladm = LADM("resources/correct-skill-ladm/skill_with_dependencies_system_only.yml")
        results = ladm.generate_dependencies_from_file()
        self.assertEqual(([],
                          ["libffi-dev", "default-jre=2:1.11-71"],
                          []), results)

    def test_libraries_only(self) -> None:
        ladm = LADM("resources/correct-skill-ladm/skill_with_dependencies_libraries_only.yml")
        results = ladm.generate_dependencies_from_file()
        self.assertEqual((["zupline==1.0.0", "tensorblow"],
                          [],
                          []), results)

    def test_correct_data(self) -> None:
        dependencies_yml_path = "resources/correct-skill-ladm/skill_with_dependencies_with_data.yml"
        ladm = LADM(dependencies_yml_path)
        results = ladm.generate_dependencies_from_file()
        self.assertEqual((["zupline==1.0.0", "tensorblow"],
                          ["libffi-dev", "default-jre=2:1.11-71"],
                          ["gaia://tenant/a.txt", "gaia://tenant/test/b.txt"]), results)

    def test_data_only(self) -> None:
        dependencies_yml_path = "resources/correct-skill-ladm/skill_with_dependencies_with_data.yml"
        ladm = LADM(dependencies_yml_path)
        results = ladm.generate_dependencies_from_file(do_language=False, do_system=False, do_data=True)
        self.assertEqual(([],
                          [],
                          ["gaia://tenant/a.txt", "gaia://tenant/test/b.txt"]), results)

    def test_planned_but_unsupported_language(self) -> None:
        dependencies_yml_path = \
            "resources/correct-skill-ladm/skill_with_dependencies_with_specified_language_java.yml"
        ladm = LADM(dependencies_yml_path)
        results = ladm.generate_dependencies_from_file()

        self.assertEqual(([],
                          ["libffi-dev", "default-jre=2:1.11-71"],
                          []), results)
        self.assertEqual(ladm.language, "java")
        self.assertEqual(ladm.version, "1.8")

    def test_unsupported_language(self) -> None:
        dependencies_yml_path = \
            "resources/incorrect-skill-ladm/skill_with_dependencies_with_specified_language_unsupported.yml"
        ladm = LADM(dependencies_yml_path)
        self.assertRaises(LADMError, ladm.generate_dependencies_from_file)

    def test_missing_language(self) -> None:
        dependencies_yml_path = \
            "resources/incorrect-skill-ladm/skill_with_dependencies_with_specified_language_missing.yml"
        ladm = LADM(dependencies_yml_path)
        self.assertRaises(LADMError, ladm.generate_dependencies_from_file)

    def test_language_version_with_single_dash(self) -> None:
        result = LADM._get_language_and_version("java-1.8")
        self.assertEqual(('java', '1.8'), result)

    def test_language_version_with_multiple_dashes(self) -> None:
        result = LADM._get_language_and_version("java-1.8-jre")
        self.assertEqual(('java', '1.8-jre'), result)



from ladm.main.generator_strategy import GeneratorStrategy
from ladm.main.util.types import DependenciesOutput, DependenciesInput


class GeneratorContext:
    def __init__(self, strategy: GeneratorStrategy = None):
        self._strategy: GeneratorStrategy = strategy

    @property
    def strategy(self) -> GeneratorStrategy:
        return self._strategy

    @strategy.setter
    def strategy(self, strategy: GeneratorStrategy) -> None:
        self._strategy = strategy

    def execute_strategy(self, dependencies: DependenciesInput) -> DependenciesOutput:
        return self._strategy.generate(dependencies)



from typing import List, Union, Hashable, Any, Dict

YamlContent = Union[Dict[Hashable, Any], List, None]

LanguageDependenciesInput = List[List[str]]
SystemDependenciesInput = List[List[str]]
DataDependenciesInput = List[str]
DependenciesInput = Union[LanguageDependenciesInput, SystemDependenciesInput, DataDependenciesInput]

LanguageDependenciesOutput = List[str]
SystemDependenciesOutput = List[str]
DataDependenciesOutput = List[str]
DependenciesOutput = Union[LanguageDependenciesOutput, SystemDependenciesOutput, DataDependenciesOutput]
